/* $Id: config.c,v 1.8 2011/04/07 14:59:17 glebius Exp $ */

#include <sys/types.h>
#include <sys/socket.h>

#include <limits.h>
#include <netdb.h>
#include <regex.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <arpa/inet.h>

#include "pfmon.h"

static FILE	*fconfig = NULL;
static char	line[LINE_MAX];

/*
 * Target statement parser:
 * 'target 10.0.0.1:80 table frontends timeout 0.5 interval 5'
 */
static char	*T_RE = "^target[ \t]+"
			"([0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}):"
			"([0-9]+)[ \t]+"
			"table[ \t]+([a-z0-9]+)[ \t]+"
			"timeout[ \t]+([0-9.]+)[ \t]+"
			"interval[ \t]+([0-9.]+)$";
#define	T_RE_NMATCH	6
static regex_t	t_re;
static regmatch_t t_match[T_RE_NMATCH];

static char *
nextline(FILE *fd)
{
	char *cp;

	if (fgets(line, sizeof (line), fd) == NULL)
		return (NULL);
	cp = strchr(line, '\n');
	if (cp)
		*cp = '\0';
	return (line);
}

static struct timespec
read_tv(char *str)
{
	struct timespec	rval;
	size_t	len;
	char	*s;

	if ((s = strchr(str, '.')) == NULL) {
		rval.tv_sec = (uint32_t )strtoul(str, (char **)NULL, 10);
		rval.tv_nsec = 0;
	} else {
		*s++ = '\0';
		len = strlen(s);
		if (len > 9)
			s[9] = '\0';
		rval.tv_sec = (long )strtoul(str, (char **)NULL, 10);
		rval.tv_nsec = (long )strtoul(s, (char **)NULL, 10);
		while (len++ < 9)
			rval.tv_nsec *= 10;
	}

	return (rval);
}

static int
settable(struct target *targ, char *tname)
{
	struct table *tab;

	SLIST_FOREACH(tab, &tabhead, link)
		if ((strncmp(tab->name, tname, sizeof(tab->name)) == 0)) {
			targ->table = tab;
			return (0);
		}

	tab = calloc(1, sizeof(*tab));
	if (tab == NULL) {
		syslog(LOG_CRIT, "calloc(): %m");
		exit(1);
	}

	strncpy(tab->name, tname, sizeof(tab->name));
	targ->table = tab;
	SLIST_INSERT_HEAD(&tabhead, tab, link);

	return (0);
}

static struct target *
getconfigent(void)
{
	struct addrinfo hint;
	struct target *targ;
	char tname[PF_TABLE_NAME_SIZE+1];
	char *cp, *iaddr, *port, *timo, *intv;
	int error;

	while ((cp = nextline(fconfig)) != NULL) {
		/* Skip comments. */
		if (*cp == '#' || *cp == '\0')
			continue;
		break;
	}
	if (cp == NULL)
		return (NULL);

	/* Match line against T_RE first. */
	if (regexec(&t_re, cp, T_RE_NMATCH, t_match, 0) != 0) {
		syslog(LOG_ERR, "Unrecognized string: %s", cp);
		return (NULL);
	}

	targ = calloc(1, sizeof(*targ));
	if (targ == NULL) {
		syslog(LOG_CRIT, "calloc(): %m");
		exit(1);
	}

	/* Insert NULLs after words. */
	iaddr = (char *)(cp + t_match[1].rm_so);
	*(char *)(cp + t_match[1].rm_eo) = '\0';
	port = (char *)(cp + t_match[2].rm_so);
	*(char *)(cp + t_match[2].rm_eo) = '\0';
	timo = (char *)(cp + t_match[4].rm_so);
	*(char *)(cp + t_match[4].rm_eo) = '\0';
	intv = (char *)(cp + t_match[5].rm_so);
	*(char *)(cp + t_match[5].rm_eo) = '\0';

	/* Read table name. */
	if ((t_match[3].rm_eo - t_match[3].rm_so) > PF_TABLE_NAME_SIZE) {
		free(targ);
		syslog(LOG_ERR, "table name too long");
		return (NULL);
	}
	strlcpy(tname, (char *)(cp + t_match[3].rm_so),
	    (t_match[3].rm_eo - t_match[3].rm_so) + 1);

	/* Read timeout values. */
	targ->timo = read_tv(timo);
	if (targ->timo.tv_sec > 60) {
		free(targ);
		syslog(LOG_ERR, "timeout is too big for %s:%s (max 60)",
		    iaddr, port);
		return (NULL);
	}
	targ->interval = read_tv(intv);
	if (targ->interval.tv_sec > 3600) {
		free(targ);
		syslog(LOG_ERR, "interval is too big for %s:%s (max 3600)",
		    iaddr, port);
		return (NULL);
	}

	/* String parsed successfully, now try to fill in addrinfo. */
	bzero(&hint, sizeof(hint));
	hint.ai_flags = AI_NUMERICHOST;
	hint.ai_family = AF_INET;
	hint.ai_socktype = SOCK_STREAM;

	error = getaddrinfo(iaddr, port, &hint, &targ->ai);
	if (error) {
		free(targ);
		syslog(LOG_ERR, "Failed getaddrinfo() for %s:%s: %s",
			iaddr, port, gai_strerror(error));
		return (NULL);
	}

	if (settable(targ, tname) != 0) {
		free(targ);
		return (NULL);
	}

	return (targ);
}

void
config(const char *cfgfile)
{
	struct target *new;
	int rval;

	if (cfgfile == NULL)
		cfgfile = PATH_CFG_FILE;

	fconfig = fopen(cfgfile, "r");
	if (fconfig == NULL) {
		syslog(LOG_ERR, "%s: %m", cfgfile);
		exit(1);
	}

	if ((rval = regcomp(&t_re, T_RE, REG_EXTENDED |
	    REG_ICASE | REG_NEWLINE)) != 0) {
		syslog(LOG_ERR, "regcomp() failed: %d", rval);
		exit(1);
	}
	SLIST_INIT(&thead);
	SLIST_INIT(&tabhead);
	while ((new = getconfigent())) {
		SLIST_INSERT_HEAD(&thead, new, link);
		syslog(LOG_DEBUG, "Got target %s:%u, table %s", AI2IP(new->ai),
			AI2PORT(new->ai), new->table->name);
	}
	regfree(&t_re);
	(void )fclose(fconfig);
	fconfig = NULL;
}
