/* $Id: main.c,v 1.16 2012/01/11 16:18:06 glebius Exp $ */

#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include <arpa/inet.h>

#include "pfmon.h"
#include "timespec.h"

struct thead_t	thead; 
struct tabhead_t tabhead; 

static void	target_set_state(struct target *, u_int);
static void	target_work(struct target *);
static struct timeval	min_timeout(void);
static void	usage(void);

fd_set rfd, wfd;
int nfds;

int daemon_flag = 1;
char *cfgfile = NULL;

int
main(int argc, char **argv)
{
	struct sigaction sa;
	struct timeval tv;
	struct timespec ts;
	struct target *targ;
	int ch, n;

	while((ch = getopt(argc, argv, "df:")) != -1)
		switch (ch) {
		case 'd':
			daemon_flag = 0;
			break;
		case 'f':
			cfgfile = optarg;
			break;
		default:
			usage();
		}

	if (daemon_flag)
		openlog(argv[0], 0, LOG_DAEMON);
	else
		openlog(argv[0], LOG_PERROR, LOG_DAEMON);

	/* Read configuration file. */
	config(cfgfile);

	/* Avoid creating zombies from pfctl/whatever. */
	(void )sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = SIG_IGN;
	(void )sigaction(SIGCHLD, &sa, (struct sigaction *)0);

	if (SLIST_EMPTY(&thead)) {
		syslog(LOG_ERR, "No targets defined, exiting");
		exit(1);
	};

	/* Initialize. */
	(void )clock_gettime(CLOCK_MONOTONIC, &ts);
	SLIST_FOREACH(targ, &thead, link) {
		targ->last = ts;
		targ->ss = SS_NOSOCKET;
	}

	/* Daemonize. */
	if (daemon_flag)
		daemon(0, 0);

	for (;;) {
		FD_ZERO(&rfd);
		FD_ZERO(&wfd);
		nfds = 0;

		SLIST_FOREACH(targ, &thead, link)
			target_work(targ);

		tv = min_timeout();
reselect:
		n = select(nfds+1, &rfd, &wfd, NULL, &tv);

		if (n == -1 && errno == EINTR)
			goto reselect;

		if (n == 0)		/* Timeout. */
			continue;

		if (n == -1)
			syslog(LOG_ERR, "select(): %m");

		SLIST_FOREACH(targ, &thead, link)
			if (targ->ss == SS_ISCONNECTING &&
			    FD_ISSET(targ->sock, &wfd))
				targ->ss = SS_CONNECTED;
	}

	return (0);
}

static void
target_set_state(struct target *targ, u_int state)
{
	if (targ->state == state)
		return;

	if (state == STATE_DOWN) {
		if (targ->state == STATE_UP)
			targ->table->count--;
		if (targ->table->count == 0)
			syslog(LOG_NOTICE, "%s:%u changed state to DOWN, "
			    "but left in %s",
			    AI2IP(targ->ai), AI2PORT(targ->ai),
			    targ->table->name);
		else {
			pfctl(CMD_DELETE, targ->table->name, AI2IP(targ->ai));
			syslog(LOG_NOTICE, "%s:%u changed state to DOWN, %s "
			    "count %d", AI2IP(targ->ai), AI2PORT(targ->ai),
			    targ->table->name, targ->table->count);
		}
	}
	if (state == STATE_UP) {
		targ->table->count++;
		pfctl(CMD_ADD, targ->table->name, AI2IP(targ->ai));
		syslog(LOG_NOTICE, "%s:%u changed state to UP, %s count %u",
		    AI2IP(targ->ai), AI2PORT(targ->ai), targ->table->name,
		    targ->table->count);
	}
	targ->state = state;
}

static void
target_work(struct target *targ)
{
	struct timespec now, ts1;
	char buf[4];
	ssize_t n;
	int error;

	(void )clock_gettime(CLOCK_MONOTONIC, &now);

start:
	switch (targ->ss) {
	case SS_NOSOCKET:
		ts1 = now;
		timespecsub(&ts1, &targ->last);
		if (timespeccmp(&ts1, &targ->interval, <))
			return;

		targ->sock = socket(targ->ai->ai_family,
	    	    targ->ai->ai_socktype, targ->ai->ai_protocol);
		if (targ->sock == -1) {
			syslog(LOG_ERR, "socket(): %m");
			exit(1);
		};
		if(fcntl(targ->sock, F_SETFL, O_NONBLOCK) == -1) {
			syslog(LOG_ERR, "fcntl(O_NONBLOCK): %m");
			exit(1);
		};
		targ->ss = SS_DISCONNECTED;
		/* fall thru */

	case SS_DISCONNECTED:
		targ->start = now;
		error = connect(targ->sock, targ->ai->ai_addr,
		    targ->ai->ai_addrlen);
		if (error == 0) {
			targ->ss = SS_CONNECTED;
			goto start;
		}

		if (errno ==  EINPROGRESS) {
			targ->ss = SS_ISCONNECTING;
			goto start;
		} else {
			syslog(LOG_ERR, "connect() to %s:%u: %m",
			    AI2IP(targ->ai), AI2PORT(targ->ai));
			target_set_state(targ, STATE_DOWN);
			close(targ->sock);
			targ->ss = SS_NOSOCKET;
			targ->last = now;
			return;
		}
		/* not reached */
		break;

	case SS_ISCONNECTING:
		ts1 = now;
		timespecsub(&ts1, &targ->start);
		if (timespeccmp(&ts1, &targ->timo, <=)) {
			FD_SET(targ->sock, &rfd);
			FD_SET(targ->sock, &wfd);
			if (targ->sock > nfds)
				nfds = targ->sock;
			return;
		} else {
			syslog(LOG_ERR, "timeout connecting to %s:%u",
			    AI2IP(targ->ai), AI2PORT(targ->ai));
			target_set_state(targ, STATE_DOWN);
			FD_CLR(targ->sock, &rfd);
			FD_CLR(targ->sock, &wfd);
			close(targ->sock);
			targ->ss = SS_NOSOCKET;
			targ->last = now;
			return;
		}
		break;

	case SS_CONNECTED:
		n = recv(targ->sock, (void *)buf, 1, MSG_PEEK);
		if (n >= 0 || errno == EAGAIN) {
			target_set_state(targ, STATE_UP);
			close(targ->sock);
			targ->ss = SS_NOSOCKET;
			targ->last = now;
		} else {
			syslog(LOG_ERR, "recv() from %s:%u: %m",
			    AI2IP(targ->ai), AI2PORT(targ->ai));
			target_set_state(targ, STATE_DOWN);
			close(targ->sock);
			targ->ss = SS_NOSOCKET;
			targ->last = now;
		}
		return;
		break;
	default:
		syslog(LOG_ERR, "Invalid targ->ss = %u", targ->ss);
	}
}

static struct timeval
min_timeout()
{
	struct timespec	now, rval, ts1;
	struct timeval tv;
	struct bintime bt;
	struct target	*targ;

	(void )clock_gettime(CLOCK_MONOTONIC, &now);
	rval.tv_sec = LONG_MAX;

	SLIST_FOREACH(targ, &thead, link) {
		/* Check interval timespec. */
		ts1 = now;
		timespecsub(&ts1, &targ->last);
		if (timespeccmp(&ts1, &rval, <))
			rval = ts1;

		/* Check connect() timeout timespec. */
		ts1 = now;
		timespecsub(&ts1, &targ->start);
		if (timespeccmp(&ts1, &rval, <))
			rval = ts1;
	}

	timespec2bintime(&rval, &bt);
	bintime2timeval(&bt, &tv);
	return (tv);
}

static void
usage()
{
	(void)fprintf(stderr, "usage: pfmon [-d] [-f config]\n");
	exit(EX_USAGE);
}
